#    builder configuration
#    Copyright (C) 2009 Mike Reed <gedit@amadron.com>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from gi.repository import Gio, GLib
import logging
import logging.config
import os.path


# The compile and build commands are determined by the directory
# in/under which the file resides.  So you can define a build/compile
# command for any directory and the deepest path for which a
# build/compile command is defined for the given file will be used.
# There will always be a build/compile command defined for "" which
# then becomes the de facto "default" commands.

class Config(object):
    """ Wraps configuration data
    
        The abstract data seen by clients of Config is a dictionary with
        keys of directory fullpath to a pair of compile command and
        build command.
        
        Currently, this is a wrapper around GSettings.  Data is placed in 
        GCONF_ROOT.  It is stored as a map of string (the root dir) keys
        to a tuple of two strings (compile cmd, build cmd). Conceptually
        I think it should map to a list of two strings but you can't a
        fixed size array in GVariant. 
    """

    GSETTINGS_ROOT = 'org.gnome.gedit.plugins.builder'
    _DFLT_ROOT = ""

    def __init__(self, data_dir):
        self._l = logging.getLogger('plugin.builder')
        self._data_dir = data_dir
        self._config_changed_cb = None
        self._commands = {}
        self._colors = {}
        self._gv_type = {}
        
        schema_source = Gio.SettingsSchemaSource.new_from_directory(
            data_dir,
            Gio.SettingsSchemaSource.get_default(),
            False
            )
        schema = Gio.SettingsSchemaSource.lookup(
            schema_source,
            self.GSETTINGS_ROOT,
            True
        )
        self._settings = Gio.Settings.new_full(schema, None, None)
        self._handler_id = self._settings.connect(
            "changed::commands", self._load)        
        self._settings.connect("changed::colors", self._load)        
        self._settings.connect("changed::message-regex", self._load)        
        self._settings.connect("changed::dir-in-regex", self._load)        
        self._settings.connect("changed::dir-out-regex", self._load)        
        self._load(self._settings, "commands")                
        self._load(self._settings, "colors")                
    
    def _load(self, settings, key):
        """ GSettings returns immutable GLib.Variant objects.  Load
            these into our own mutable objects
        """
        self._l.debug("Entered")
        self._l.info("Loading [{}]".format(key))
        if key == "commands":
            self._load_map(settings, key, self._commands)
            if self._DFLT_ROOT not in self._commands.keys():
                self._l.info("No default command in GSettings. Adding...")
                self._commands[self._DFLT_ROOT] = ('make %s.o', 'make')
                self._settings.disconnect(self._handler_id)        
                self._store(key)
                self._handler_id = self._settings.connect(
                    "changed::commands", self._load)
        elif key == "colors":
            self._load_map(settings, key, self._colors)
            if len(self._colors) == 0:
                self._l.info("No colors in GSettings.  Adding...")
                # Setup default values
                self._colors = {
                    'error':'#DF2424',
                    'info':'#4242CD',
                    'warning':'#F5A20C',
                    'note': '#4242CD',
                    'default':'#4242CD',
                }
        elif key == "message-regex":
            self._message_regex = settings.get_value("message-regex")
        elif key == "dir-in-regex":
            self._dir_in_regex = settings.get_value("dir-in-regex")
        elif key == "dir-out-regex":
            self._dir_out_regex = settings.get_value("dir-out-regex")
            
        if self._config_changed_cb:
            self._config_changed_cb()
        self._l.debug("Exiting")
    
    def _load_map(self, settings, key, target):
        self._l.debug("Entered: {}".format(target))
        target.clear()
        gs_variant = settings.get_value(key)
        self._gv_type[key] = gs_variant.get_type_string()
        for k in gs_variant.keys():
            self._l.debug("{}={}".format(k, gs_variant[k]))
            target[k] = gs_variant[k]
        self._l.debug("Exiting")
     
    def _store(self, key):
        self._l.debug("Entered: saving {}".format(key))
        if key == "commands":
            self._settings.set_value(
                key,
                GLib.Variant(self._gv_type[key], self._commands))
        elif key == "colors":
            self._settings.set_value(
                key,
                GLib.Variant(self._gv_type[key], self._colors))
        self._l.debug("Exiting")

    def _get_rootdir(self, directory):
        dir_len = len(directory)
        longest = self._DFLT_ROOT
        for rd in self._commands.keys():
            rd_len = len(rd)
            if (rd_len > len(longest) and
                rd_len <= dir_len and
                directory[:rd_len] == rd):
                longest = rd
        return longest
        
    def get_data_dir(self):
        """ The plugin directory
        """
        return self._data_dir
        
    def set_config_change_cb(self, cb):
        self._config_changed_cb = cb

    def get_rootdirs(self):
        """ A list of all root dirs for which commands have been specified
        """
        return self._commands.keys()
        
    def get_commands(self, directory):
        """ Returns a tuple of (compile cmd, build cmd) for a directory
        """
        rd = self._get_rootdir(directory)
        self._l.info("Root dir for [{}] is [{}]".format(directory, rd))
        return self._commands[rd]

    def del_commands(self, rd):
        if rd in self._commands:
            del self._commands[rd]
            self._store("commands")
        
    def save_commands(self, rd, c, b):
        if type(c) is str and type(b) is str:
            self._commands[rd] = (c, b)
            self._store("commands")
        
    def compile_cmd(self, uri):
        cmd = self.get_commands(uri)[0]
        self._l.info('Compiling [%s]  with [%s]' % \
            (uri, cmd))
        return _subst(uri, cmd)

    def build_cmd(self, uri):
        cmd = self.get_commands(uri)[1]
        self._l.info('Building [%s] with [%s]' % \
            (uri, cmd))
        return _subst(uri, cmd)

    def build_root(self, uri):
        rd = self._get_rootdir(uri)
        self._l.info('build dir for [%s] is [%s]' % \
                         (uri, rd))
        if rd == self._DFLT_ROOT:
            # For the default commands use the CWD as the build root
            return _subst(uri, "%d")
        return rd
    
    def severities(self):
        return self._colors
        
    def save_severity(self, sev, color):
        self._l.debug("Entered: {}:{}".format(sev, color))
        self._colors[sev] = color
        self._store("colors")
        self._l.debug("Exiting")
        
    def del_severity(self, sev):
        self._l.debug("Entered: {}".format(sev))
        del(self._colors[sev])
        self._store("colors")
        self._l.debug("Exiting")
        
    def message_regex(self):
        return self._settings.get_value('message-regex').get_string().replace('\n', '')
        
    def save_message_regex(self, re):
        self._settings.set_value('message-regex', GLib.Variant('s', re))
        
    def dir_in_regex(self):
        return self._settings.get_value('dir-in-regex').get_string().replace('\n', '')
        
    def save_dir_in_regex(self, re):
        self._settings.set_value('dir-in-regex', GLib.Variant('s', re))
        
    def dir_out_regex(self):
        return self._settings.get_value('dir-out-regex').get_string().replace('\n', '')
        
    def save_dir_out_regex(self, re):
        self._settings.set_value('dir-out-regex', GLib.Variant('s', re))
        
def _subst(fullpath, cmd):
    (directory, filename) = os.path.split(fullpath)
    (stem, extension) = os.path.splitext(filename)
    if len(extension) > 0: extension = extension[1:]
    cmd = cmd.replace('%%', '\0\0')
    cmd = cmd.replace('%s', stem)
    cmd = cmd.replace('%e', extension)
    cmd = cmd.replace('%f', filename)
    cmd = cmd.replace('%d', directory)
    cmd = cmd.replace('%p', fullpath)
    cmd = cmd.replace('\0\0', '%')
    return cmd
        

if __name__ == "__main__":
    import unittest
    
    class Tests(unittest.TestCase):
        DATA_DIR = "."
    
        def setUp(self):
            # Save current GSettings
            schema_source = Gio.SettingsSchemaSource.new_from_directory(
                self.DATA_DIR,
                Gio.SettingsSchemaSource.get_default(),
                False
                )
            schema = Gio.SettingsSchemaSource.lookup(
                schema_source,
                Config.GSETTINGS_ROOT,
                True
            )
            self.settings = Gio.Settings.new_full(schema, None, None)
            self.prev = self.settings.get_value("commands")
            
            # Setup test GSettings
            c = {}
            c["/a1/a2/a3"] = ("a3c %f", "a3b")
            c["/a1/a2/a3/a4/a5"] = ("a5c %f", "a5b")
            c["/a1/a2"] = ("a2c %f", "a2b")
            self.settings.set_value("commands", GLib.Variant("a{s(ss)}", c))

            self.c = Config(self.DATA_DIR)
            print(self.c.message_regex())
            
        def tearDown(self):
            # Restore previous settings
            self.settings.set_value("commands", self.prev)
            
        def test_data_dir(self):
            self.assertEqual(self.c.get_data_dir(), self.DATA_DIR)
            
        def test_subst(self):
            self.assertEqual(
                _subst(
                    "/top/sub/stem.ext",
                    "doublepercent=%% stem=%s ext=%e file=%f dir=%d path=%p"),
                "doublepercent=% "
                "stem=stem "
                "ext=ext "
                "file=stem.ext "
                "dir=/top/sub "
                "path=/top/sub/stem.ext"
            )

        def testget_commands(self):
            cmds = self.c.get_commands("rubbish")
            self.assertEqual(cmds[0], "make %s.o")
            self.assertEqual(cmds[1], "make")

            cmds = self.c.get_commands("/a1")
            self.assertEqual(cmds[0], "make %s.o")
            self.assertEqual(cmds[1], "make")

            cmds = self.c.get_commands("/a1/a2")
            self.assertEqual(cmds[0], "a2c %f")
            self.assertEqual(cmds[1], "a2b")

            cmds = self.c.get_commands("/a1/a2/b3/b4")
            self.assertEqual(cmds[0], "a2c %f")
            self.assertEqual(cmds[1], "a2b")

            cmds = self.c.get_commands("/a1/a2/a3/b4")
            self.assertEqual(cmds[0], "a3c %f")
            self.assertEqual(cmds[1], "a3b")

            cmds = self.c.get_commands("/a1/a2/a3/a4")
            self.assertEqual(cmds[0], "a3c %f")
            self.assertEqual(cmds[1], "a3b")

            cmds = self.c.get_commands("/a1/a2/a3/a4/")
            self.assertEqual(cmds[0], "a3c %f")
            self.assertEqual(cmds[1], "a3b")

            cmds = self.c.get_commands("/a1/a2/a3/a4/a5")
            self.assertEqual(cmds[0], "a5c %f")
            self.assertEqual(cmds[1], "a5b")

            cmds = self.c.get_commands("/a1/b1")
            self.assertEqual(cmds[0], "make %s.o")
            self.assertEqual(cmds[1], "make")

            cmds = self.c.get_commands("")
            self.assertEqual(cmds[0], "make %s.o")
            self.assertEqual(cmds[1], "make")

        def test_(self):
            pass
            
    unittest.main()



