#    builder configuration editor
#    Copyright (C) 2009 Mike Reed <gedit@amadron.com>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


import logging
import os.path
from gi.repository import Gtk, Gdk

class ConfigEditor(object):
    _NEW_PATH_ENTRY = ' <edit new path>'
    _DFLT_TEXT = ' <default>'
    
    def __init__(self, config):
        self._l = logging.getLogger('plugin.builder')
        self._l.debug('Entered: data_dir[%s]' % config.get_data_dir())
        self._ui = None
        self._glade_file = os.path.join(config.get_data_dir(), 'ConfigEditor.glade')
        self._config = config
        
    def dialog(self):
        self._l.debug('Entered')

        if not self._ui:
            self._l.info('Creating new dialog')

            ui_builder = Gtk.Builder()
            ui_builder.add_from_file(self._glade_file)
            self._ui = ui_builder.get_object
            
            ui_builder.connect_signals(self)
            #self._ui('config_root_path').set_sort_indicator(True)
            self._ui('config_root_path').set_sort_column_id(0)
            self._ui('config_root_path').clicked()
            self._ui('config_rootdirs').set_sort_func(0, self._sort)
                
        self._populate()

        # select the first row (which should always be "").  This will
        # trigger the CB and populate the compile/build entries
        sel = self._ui('config_listview').get_selection()
        self._current_rootdir = None
        self._current_cmp = None
        self._current_bld = None
        sel.select_path(Gtk.TreePath('0'))
        
        self._l.debug('Exiting')
        return self._ui('config_editor')
    
    def _sort(self, model, i1, i2, user_data):
        """ Specify how the rootdirs sort
            For some reason the default is odd and doesn't seem
            to have standard ASCII ordering with symbols like /
            <spc> < and (.
        """
        v1 = model.get_value(i1, 0)
        v2 = model.get_value(i2, 0)
        self._l.debug("Comparing [{}] and [{}]".format(v1, v2))
        return cmp(v1, v2)
    
    def _populate(self):
        """ Take info from Config and put it in the UI
        """
        self._l.debug('Entered')
        list_store = self._ui('config_rootdirs')
        list_store.clear()
        list_store.append((self._NEW_PATH_ENTRY, True))
        rds = self._config.get_rootdirs()
        for d in sorted(rds):
            if d == '':
                list_store.append([self._DFLT_TEXT, False])
            else:
                list_store.append([d, True])

        list_store = self._ui('config_colors')
        list_store.clear()
        for sev, color in self._config.severities().iteritems():
            list_store.append([sev, color])

    def _save_edits(self):
        """ Check for changes in the cmp and bld commands.
            If the commands have changed save the changes
            back to Config
        """
        self._l.debug('Entered')
        if self._current_rootdir is None:
            return
            
        text_cmp = self._ui('config_compile').get_text()
        text_bld = self._ui('config_build').get_text()
        if (self._current_cmp != text_cmp or
            self._current_bld != text_bld):
            self._l.info('cmp[{} -> {}]'.format(self._current_cmp, text_cmp))
            self._l.info('bld[{} -> {}]'.format(self._current_bld, text_bld))
            if self._current_rootdir == self._DFLT_TEXT:
                config_rootdir = ''
            else:
                config_rootdir = self._current_rootdir
            self._config.save_commands(
                config_rootdir,
                text_cmp,
                text_bld)
        self._l.debug('Exiting')
        
    def _listsel_changed_cb(self, sel):
        self._l.debug('Entered')
        
        self._save_edits()
        
        # Now update text boxes with commands for newly
        # selected root dir
        (list_store, sel_row) = sel.get_selected()
        if sel_row:
            new_rd = list_store.get_value(sel_row, 0)
            
            if new_rd != self._current_rootdir:
                self._l.debug("["+new_rd+"]")
                if new_rd == self._NEW_PATH_ENTRY:
                    new_cmp = ''
                    new_bld = ''
                    editable = False
                else:
                    new_cmp, new_bld = self._config.get_commands(new_rd)
                    editable = True
                self._ui('config_compile').set_text(new_cmp)
                self._ui('config_compile').set_sensitive(editable)
                self._ui('config_build').set_text(new_bld)
                self._ui('config_build').set_sensitive(editable)
                self._current_rootdir = new_rd
                self._current_cmp = new_cmp
                self._current_bld = new_bld
        self._l.debug('Exiting')
        
    def _path_edited_cb(self, rend, sel_path, new_text):
        self._l.debug('Entered')
        self._l.info('Edited item is {}'.format(sel_path))
        self._l.info('New path is [{}]'.format(new_text))

        list_store = self._ui('config_rootdirs')
        sel_iter = list_store.get_iter(Gtk.TreePath(sel_path))
        old_rd = list_store.get_value(sel_iter, 0)

        if len(new_text) == 0:
            if old_rd != self._NEW_PATH_ENTRY:
                # delete the entry
                self._l.info('Deleting {}'.format(old_rd))
                self._current_rootdir = None
                del list_store[sel_iter]
                self._config.del_commands(old_rd)
        elif new_text != old_rd and self._path_ok(new_text):
            if new_text != old_rd:
                self._l.info("rootdir[{} -> {}]".format(old_rd, new_text))
                
                list_store.set_value(sel_iter, 0, new_text)
                if old_rd == self._NEW_PATH_ENTRY:
                    list_store.insert_before(sel_iter, [self._NEW_PATH_ENTRY, True])
                    self._config.save_commands(new_text, '', '')
                    self._ui('config_listview').scroll_to_cell(
                        list_store.get_path(sel_iter)
                    )    
                    self._ui('config_compile').set_sensitive(True)
                    self._ui('config_build').set_sensitive(True)
                else:
                    self._config.del_commands(old_rd)
                    self._config.save_commands(
                        new_text, self._current_cmp, self._current_bld)
                self._current_rootdir = new_text
        
        self._l.debug('Exiting')
        
    def _path_ok(self, path):
        """ Check the rootdir path entered by user
        """
        error_message = None
        if len(path) == 0 or path[0] != '/':
            error_message = 'Path must be absolute i.e. it must start with /'
        elif path in self._config.get_rootdirs():
            error_message = 'Path already exists'
        
        if error_message:
            md = Gtk.MessageDialog(
                self._ui('config_editor').get_parent().get_parent(),
                Gtk.DialogFlags.MODAL,
                Gtk.MessageType.ERROR,
                Gtk.ButtonsType.OK,
                error_message)
            md.run()
            md.destroy()

        return error_message is None
    
    def _choose_colour_cb(self, treeview, event):
        self._l.debug("Entered")

        info = treeview.get_path_at_pos(event.x, event.y)
        if info is None:
            self._l.debug("No row clicked on.  Exiting")
            return False
            
        path, column, cellx, celly = info
        model = treeview.get_model()
        sel_iter = model.get_iter(path)
        color_string = model.get_value(sel_iter, 1)
        
        if column.get_title() == 'Color':
            csd = Gtk.ColorSelectionDialog('Pick Colour')
            cs = csd.get_color_selection()
            cs.set_has_opacity_control(False)
            cs.set_previous_color(Gdk.color_parse(color_string))
            cs.set_current_color(Gdk.color_parse(color_string))
            if csd.run() == Gtk.ResponseType.OK:
                color_string = cs.get_current_color().to_string()
                model.set_value(sel_iter, 1, color_string)
                self._l.debug(color_string)
                self._config.save_severity(model.get_value(sel_iter, 0), color_string)
            csd.destroy()
        else:
            treeview.get_selection().set_mode(Gtk.SelectionMode.SINGLE)
            treeview.set_cursor_on_cell(path, column, None, True)

        return True
        self._l.debug("Exiting")
        
    def _sev_edited_cb(self, rend, sel_path, new_text):
        self._l.debug('Entered')
        self._l.debug(sel_path)
        self._l.debug(new_text)
        self._l.debug(rend.get_property('editing'))
        cc = self._ui('config_colors')
        prev_text = cc.get_value(cc.get_iter(sel_path), 0)
        if new_text != prev_text:
            color_text = cc.get_value(cc.get_iter(sel_path), 1)
            self._config.del_severity(prev_text)
            self._config.save_severity(new_text, color_text)
        
        self._ui('config_listview_colors').get_selection().set_mode(Gtk.SelectionMode.NONE)
        self._l.debug('Exiting')
    
    def _sev_edited_canceled_cb(self, rend):
        self._l.debug('Entered')
        self._ui('config_listview_colors').get_selection().set_mode(Gtk.SelectionMode.NONE)
        self._l.debug('Exiting')
    
    def _editor_destroy_cb(self, dialog):
        self._l.debug('Entered')
        self._save_edits()
        self._ui = None
        self._l.debug('Exiting')
        

# UNIT TESTS.  As this is a GUI just bring up the ConfigEditor in a Window
if __name__ == '__main__':
    from Config import Config

    f = logging.Formatter(
        '%(asctime)s: %(levelname)8s: %(module)s(%(lineno)d): %(funcName)s: %(message)s')
    h = logging.StreamHandler()
    h.setLevel(logging.DEBUG)
    h.setFormatter(f)
    l = logging.getLogger("plugin.builder")
    l.addHandler(h)
    l.setLevel(logging.DEBUG)

    dialog = Gtk.Dialog('Test ConfigEditor')
    cwd = os.path.dirname(os.path.realpath(__file__))
    cfg = Config(cwd)
    ce = ConfigEditor(cfg)
    dialog.get_content_area().add(ce.dialog())
    dialog.add_button(Gtk.STOCK_CLOSE, 1)
    dialog.run()
    dialog.destroy()
    
