# -*- coding: utf-8 -*-

import logging
# logging.DEBUG on dev, logging.WARNING on prod
logging.basicConfig()

from . import utils

# taken from bzrpanel
class GeditPluginHelper(object):

    def __init__(self):
        self._handlers = {}
        self._htogglers = {}
        self._log = logging.getLogger(self.__class__.__name__)
        self._log.setLevel(logging.DEBUG)
        
    def _connect(self, gobj, signal, func, *args):
        handlers = self._handlers.setdefault(gobj, [])
        handlers.append(gobj.connect(signal, func, *args))
        
    def _connect_after(self, gobj, signal, func, *args):
        handlers = self._handlers.setdefault(gobj, [])
        handlers.append(gobj.connect_after(signal, func, *args))
    
    def _toggle_handlers(self, gobj):
        '''Bloquer / débloquer des écouteurs d'événements'''
        try:
            handlers = self._handlers[gobj]
        except KeyError:
            return
        toggle = self._htogglers.setdefault(gobj, False)
        for lid in handlers:
            if toggle:
                gobj.handler_block(lid)
            else:
                gobj.handler_unblock(lid)
        self._htogglers[gobj] = not toggle
    
    def _disconnect(self, gobj):
        '''Deconnecter les écouteurs sur un objet.'''
        try:
            handlers = self._handlers[gobj]
        except KeyError:
            return
        while handlers:
            gobj.disconnect(handlers.pop())
        del self._handlers[gobj]
        
    def _disconnect_all(self):
        '''Deconnecter tous les écouteurs enregistrés.'''
        for gobj, handlers in self._handlers.items():
            while handlers:
                gobj.disconnect(handlers.pop())
        self._handlers.clear()


class GeditPluginSettings(object):
    '''Objet permettant de gérer les préférences d'un plugin Gedit'''
    
    def __init__(self, current_dir, plugin_name):
        self._settings = utils.get_settings(current_dir, plugin_name)
    
    def connect_change(self, key, listener):
        self._settings.connect('changed::' + key, listener)
        
    def get_string(self, key):
        '''Récupérer une préférence de type chaîne de caractères
            @param {String} key Nom de la préférence
            @return {String}
        '''
        return self._settings.get_string(key)
        
    def set_string(self, key, value):
        ''' Définir une préférence de type chaîne de caractères.
            @param {String} key Nom de la préférence.
            @param {String} value Valeur de la préférence.
        '''
        self._settings.set_string(key, value)

    def get_boolean(self, key):
        '''Récupérer une préférence de type booléen.
            @param {String} key Nom de la préférence
            @return {Boolean}
        '''
        return self._settings.get_boolean_value(key)
        
    def set_boolean(self, key, value):
        ''' Définir une préférence de type booléen.
            @param {String} key Nom de la préférence.
            @param {Boolean} value Valeur de la préférence.
        '''
        self._settings.set_boolean_value(key, value)


class GeditPluginConfig(object):
    '''Objet permettant de gérer la fenètre de configuration d'un plugin Gedit'''

    def __init__(self, app_name, current_dir, settings):
        self._builder = utils.get_gtk_builder(current_dir, 'config.glade', ['config'], app_name)
        self._widget = None
        self._settings = settings
        
    def get_widget(self):
        ''' Récupérer le contenu du panneau de configuration.
            @return GtkWidget
        '''
        self._widget = self._builder.get_object('config')
        self._update_widget()
        self._builder.connect_signals(self)
        
        return self._widget
        
    def _update_widget(self):
        ''' Méthode appelée dans get_widget pour assurer la construction.'''
        pass

    def _map_widgets(self, prefnames):
        for name in prefnames:
            widget = self._builder.get_object(name)
            if widget is not None:
                widget.set_data('prefname', name)
               
    def on_action(self, actionable, data=None):
        if actionable.get_active():
            prefname = actionable.get_data('prefname')
            self._settings.set_string(prefname, actionable.get_action_name())
        
    def on_boolean_value_toggled(self, button, data=None):
        ''' '''
        prefname = button.get_data('prefname')
        self._settings.set_boolean(prefname, prefname.get_active())
    
    def on_string_value_changed(self, editable, data=None):
        ''' '''
        prefname = editable.get_data('prefname')
        self._settings.set_string(prefname, editable.get_text())

